import pandas as pd
import numpy as np
from sklearn.neighbors import KNeighborsClassifier
from scipy import interpolate
import scipy.optimize

def calculate_fractions(df):
    df['Fraction Pale'] = np.divide(df['Area Pale'],df['Area']-df['Area Old Mortality'])
    df['Fraction Partially Bleached'] = np.divide(df['Area Partial Bleaching'],df['Area']-df['Area Old Mortality'])
    df['Fraction Bleached'] = np.divide(df['Area Bleached'],df['Area']-df['Area Old Mortality'])
    df['Fraction Bleaching'] = np.divide(df['Area Bleach Affected'],df['Area']-df['Area Old Mortality'])
    df['Fraction Recent Mortality'] = np.divide(df['Area Recent Mortality'],df['Area']-df['Area Old Mortality'])
    df['Fraction Diseased'] = np.divide(df['Area Diseased'],df['Area']-df['Area Old Mortality'])
    df['Fraction Unhealthy'] = np.divide(df['Area Unhealthy'],df['Area']-df['Area Old Mortality'])
    df['Fraction Total Mortality'] = np.divide(df['Area Old Mortality']+ df['Area Recent Mortality'],df['Area'])
    if ('Area Disease Recent Mortality' in df.columns):
        df['Fraction Disease Mortality'] = np.divide(df['Area Disease Recent Mortality'],df['Area']-df['Area Old Mortality'])
        df.loc[np.isnan(df['Area Disease Recent Mortality']),'Fraction Disease Mortality'] = np.nan
    if ('Area Other Recent Mortality' in df.columns):
        df['Fraction Disease Mortality'] = np.divide(df['Area Other Recent Mortality'],df['Area']-df['Area Old Mortality'])
        df.loc[np.isnan(df['Area Other Recent Mortality']),'Fraction Other Mortality'] = np.nan
    if ('Area Disease Recent Mortality' in df.columns):
        df['Fraction Disease Mortality'] = np.divide(df['Area Disease Recent Mortality'],df['Area']-df['Area Old Mortality'])
    if ('Area DAB' in df.columns):
        df['Fraction Sedimented'] = np.divide(df['Area Sedimented'],df['Area']-df['Area Old Mortality'])
        df['Fraction DAB'] = np.divide(df['Area DAB'],df['Area']-df['Area Old Mortality'])
        df['Fraction DAB Mortality'] = np.divide(df['Area DAB Mortality'],df['Area']-df['Area Old Mortality'])
        df['Fraction DAB Unhealthy'] = np.divide(df['Area DAB Unhealthy'],df['Area']-df['Area Old Mortality'])
    return df
        
def read_ncrmp_data(path_to_csv):
    frrp_data = pd.read_csv(path_to_csv)
    frrp_data['Area'] = np.multiply(frrp_data['Width'],frrp_data['Width'])
    frrp_data['Area Old Mortality'] = np.multiply(frrp_data['Area'], frrp_data['Old'] / 100)
    frrp_data['Area Pale'] = np.multiply(frrp_data['Area']-frrp_data['Area Old Mortality'], frrp_data['Bleaching'] == 'P')
    frrp_data['Area Partial Bleaching'] = np.multiply(frrp_data['Area']-frrp_data['Area Old Mortality'], frrp_data['Bleaching'] == 'PB')
    frrp_data['Area Bleached'] = np.multiply(frrp_data['Area']-frrp_data['Area Old Mortality'], frrp_data['Bleaching'] == 'BL')
    frrp_data['Area Bleach Affected'] = (1/3)*frrp_data['Area Pale'] + frrp_data['Area Bleached'] + (2/3)*frrp_data['Area Partial Bleaching']
    frrp_data['Area Recent Mortality'] = np.multiply(frrp_data['Area'], (np.nanmax([frrp_data['Recent'],frrp_data['oPercRecMort']+frrp_data['dPercRecMort']],axis=0)) / 100)
    frrp_data['Area Diseased'] = np.multiply(frrp_data['Area']-frrp_data['Area Old Mortality'], np.logical_or(np.logical_not(pd.isnull(frrp_data['Disease Conditions'])),np.logical_and(np.logical_not(pd.isnull(frrp_data['dPercRecMort'])), frrp_data['dPercRecMort']> 0 )))
    frrp_data['Area Other Recent Mortality'] = np.multiply(frrp_data['Area'], frrp_data['oPercRecMort'] / 100)
    frrp_data.loc[pd.isnull(frrp_data['oPercRecMort']),'Area Other Recent Mortality'] = np.nan
    frrp_data.loc[np.logical_and(pd.isnull(frrp_data['Disease Conditions']),pd.isnull(frrp_data['dPercRecMort'])),'Area Disease Recent Mortality'] = frrp_data['Area Recent Mortality']
    frrp_data['Area Disease Recent Mortality'] = np.multiply(frrp_data['Area'], frrp_data['dPercRecMort'] / 100)
    frrp_data.loc[pd.isnull(frrp_data['dPercRecMort']),'Area Disease Recent Mortality'] = np.nan
    frrp_data.loc[np.logical_and(np.logical_not(pd.isnull(frrp_data['Disease Conditions'])),pd.isnull(frrp_data['dPercRecMort'])),'Area Disease Recent Mortality'] = frrp_data['Area Recent Mortality']
    frrp_data['Area Unhealthy'] = frrp_data.transpose().apply(lambda x: max(x['Area Diseased'],x['Area Bleach Affected']))
    df = frrp_data.groupby(['Site','Date','Zone','Subregion','Habitat']).agg(np.nanmean)
    df_other = frrp_data.drop(columns=['Bleaching','Species','TLRate','Disease Conditions', 'Other Conditions',
                                      'Width','Transect', 'Height','Old','Recent','oPercRecMort','dPercRecMort']).groupby(['Site','Date','Zone','Subregion','Habitat']).agg(np.nansum)
    df_other['Latitude'] = df['Latitude']
    df_other['Longitude'] = df['Longitude']
    df_other['Depth'] = df['Depth']
    df = df.reset_index().set_index('Site')
    df_other = df_other.reset_index().set_index('Site')
    df_other.loc[np.logical_or(df_other['Subregion']=='Deerfield',df['Subregion']=='Broward-Miami'),'Subregion'] = 'Broward'
    df_other.loc[df_other['Subregion'].apply(lambda x: 'Palm' in x),'Subregion'] = 'Palm Beach'
    df_other.loc[df_other['Subregion']=='Biscayne','Subregion'] = 'Upper Keys'
    df_other = calculate_fractions(df_other)
    df_other['Fraction Disease Mortality'] = np.divide(df_other['Area Disease Recent Mortality'],df_other['Area']-df_other['Area Old Mortality'])
    df_other.loc[df_other['Area Disease Recent Mortality']<1e-5,'Area Disease Recent Mortality'] = np.nan
    df_other.loc[np.isnan(df['Area Disease Recent Mortality']),'Fraction Disease Mortality'] = np.nan
    df_other['Fraction Other Mortality'] = np.divide(df_other['Area Other Recent Mortality'],df_other['Area']-df_other['Area Old Mortality'])
    df_other.loc[np.isnan(df['Area Other Recent Mortality']),'Fraction Other Mortality'] = np.nan
    df_other.loc[df_other['Area Other Recent Mortality']<1e-5,'Area Other Recent Mortality'] = np.nan

    df_other['Date'] = pd.to_datetime(df_other['Date'])
    frrp_data['Date'] = pd.to_datetime(frrp_data['Date'])
    return frrp_data, df_other

def read_irma_survey(path_to_csv, site_data_region_predictor=None):
    irma_data = pd.read_csv(path_to_csv)
    df = irma_data.copy()
    df['Species'] = df['Species'].apply(lambda x: x.upper())
    df['Area'] = np.multiply(df['Max_Diameter'],df['Max_Diameter'])
    df['Area Old Mortality'] = np.multiply(df['Area'], df['Old_Mortality'] / 100)
    df['Area Pale'] = np.multiply(df['Area']-df['Area Old Mortality'], df['Pale'])
    df['Area Partial Bleaching'] = np.multiply(df['Area']-df['Area Old Mortality'], df['Partial_Bleaching'])
    df['Area Bleached'] = np.multiply(df['Area']-df['Area Old Mortality'], df['Bleached'])
    df['Area Bleach Affected'] = np.multiply(df['Area']-df['Area Old Mortality'], np.logical_or.reduce(df[['Pale','Partial_Bleaching','Bleached']].transpose()))
    df['Area Recent Mortality'] = np.multiply(df['Area'], df['Recent_Mortality'] / 100)
    df['Area Broken'] = np.multiply(df['Area']-df['Area Old Mortality'],df['Broken'])
    df['Area Abrased'] = np.multiply(df['Area']-df['Area Old Mortality'],df['Abrasion'])
    df['Area Dislodged'] = np.multiply(df['Area']-df['Area Old Mortality'],df['Dislodged'])
    df['Area Sedimented'] = np.multiply(df['Area']-df['Area Old Mortality'],df['Sediment_Cover'])
    df['Area DAB'] = (df['Area']-df['Area Old Mortality'])*np.logical_or.reduce(df[['Broken','Abrasion','Dislodged']].transpose())
    df['Area DAB Mortality'] = df['Area Recent Mortality']*np.logical_or.reduce(df[['Broken','Abrasion','Dislodged']].transpose())
    df['Area Diseased'] = np.multiply(df['Area']-df['Area Old Mortality'], np.logical_or.reduce(df[['UK','DS','BB','YB','WB','WP','WS']].transpose()))
    df['Area Unhealthy'] = df.transpose().apply(lambda x: max(x['Area Diseased'],x['Area Bleach Affected']))
    df['Area DAB Unhealthy'] = df.transpose().apply(lambda x: min(x['Area Unhealthy'],x['Area DAB']))
    df['Area Disease Recent Mortality'] = np.nan
    df.loc[df['Area Diseased']>0,'Area Disease Recent Mortality'] = df['Area Recent Mortality']
    df_new = df.groupby(['Site_Code','Type']).mean()
    df_other = df.drop(columns=['Partial_Bleaching','Pale','Species','Transect_NR', 'UK','DS','BB','YB','WB','WP','WS',
                                'Abrasion','Dislodged','Broken','Sediment_Cover','Predation',
                                'Max_Diameter', 'Height','Recent_Mortality','Old_Mortality',
                                'Depth_ft','Meters_Completed','Deep Wavespeed','Shallow Wavespeed']).groupby(['Site_Code','Type']).sum()
    df_other['Latitude'] = df_new['Latitude']
    df_other['Longitude'] = df_new['Longitude']
    df_other['Depth'] = df_new['Depth_ft']
    df_other['Mean Wavespeed'] = df_new['Mean Wavespeed']
    irma_df = df
    irma_df['Date'] = pd.to_datetime('12/22/2017')
    irma_df['Depth'] = irma_df['Depth_ft']
    irma_df['Width'] = irma_df['Max_Diameter']
    irma_df['Recent'] = irma_df['Recent_Mortality']
    irma_df['Old'] = irma_df['Old_Mortality']
    irma_df['Site'] = irma_df['Site_Code'].drop(columns=['Depth_ft','Type','Site_Code','Max_Diameter','Old_Mortality','Recent_Mortality'])
    
    df_other = df_other.reset_index()
    df_other['Site'] = df_other['Site_Code']
    df_other['Zone'] = df_other['Type']

    if (site_data_region_predictor != None) :
        irma_df[['Zone','Subregion','Habitat']] = site_data_region_predictor.predict(irma_df[['Latitude','Longitude','Depth']])
        df_other[['Zone','Subregion','Habitat']] = site_data_region_predictor.predict(df_other[['Latitude','Longitude','Depth']])
    df_other = df_other.set_index('Site').drop(columns=['Site_Code','Type'])
    df_other['Date'] = pd.to_datetime('12/22/2017')
    df_other = calculate_fractions(df_other)
    return irma_df, df_other

def circle_dist(lon1,lat1,lon2,lat2):
    R = 6373.0

    lat1 = np.radians(lat1)
    lon1 = np.radians(lon1)
    lat2 = np.radians(lat2)
    lon2 = np.radians(lon2)

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2)**2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    distance = R * c
    return distance

tortugas = [-82.8738213, 24.6268887]
coords = np.array([[24.520197, -82.106816],
    [24.528778432394976, -81.98569959448399],
                   [24.543516142276374, -81.924619922095],
                   [24.547549867447668, -81.86407390591897],
                   [24.554813, -81.753093],[24.594755, -81.604682],
                   [24.62857, -81.478259],[24.62857, -81.384815],
                   [24.640267, -81.340842],[24.6709693, -81.236405],
                   [24.701842, -81.162175],[24.711618, -81.093969],[24.737813, -80.994341],
                   [24.787071, -80.890396],[24.820729, -80.815847],
                   [24.834439, -80.763629],[24.879920, -80.690454],
                   [24.947177, -80.595980],[25.009108, -80.515935],
                   [25.082514, -80.438294],[25.158361, -80.368211],
                   [25.235404, -80.333894],[25.296879, -80.291982],
                   [25.394615, -80.239077],[25.420799,-80.210659],
                   [25.686323,-80.1784745],
                   [25.8148849,-80.142769],[25.94701,-80.13933],[26.106740,-80.1235429],
                   [26.306268,-80.08989],[26.504224,-80.06517],[26.661346,-80.06105]])

def project_to_florida_coast(site_data_full):
    coastline = interpolate.interp1d(coords[:,1],coords[:,0],fill_value='extrapolate')
    closest_points = list()
    for i in site_data_full.index:
        point = np.array([site_data_full.loc[i,'Longitude'],site_data_full.loc[i,'Latitude']])
            
        if (len(point.shape)>1):
            point = point[:,0]
        if (point[0] < -82.5 or ('Subregion' in site_data_full.columns and 'Tortugas' in site_data_full.loc[i,'Subregion'])):
            site_data_full.loc[i, 'Distance to Shore'] = np.linalg.norm(tortugas - point)
            closest_points.append(tortugas)
            continue
            
        nearest = scipy.optimize.minimize(lambda x : circle_dist(x,coastline(x),point[0],point[1]), point[0],method = 'Nelder-Mead',tol=1e-7)
        nearest = nearest.x[0]
        
        if (nearest < -82.3):
            nearest = -82
        nearest = [nearest, coastline(nearest)]
        
        if (nearest[1] + 0.15 < point[1]):
            site_data_full.loc[i, 'Distance to Shore'] = np.nan
            closest_points.append(nearest)
            continue
        site_data_full.loc[i, 'Distance to Shore'] = circle_dist(nearest[0],nearest[1],point[0],point[1])
        closest_points.append(nearest)
    site_data_full['Shore Projection'] = closest_points
    return site_data_full

def calc_hurricane_distance_windspeeds(df, start, end, intensity, rmw):
    hurr_dis = end - start
    windspeeds = np.zeros(len(df))
    distances = np.zeros(len(df))
    for idx, i in enumerate(df.index):
        point = df.loc[i,['Longitude','Latitude']]
        if (len(point.shape)>1):
            point = point.iloc[0]
        if (all(start == end)):
            d = 111*np.linalg.norm(point - start)
        else:
            proj_point = np.dot(point - start,hurr_dis)/np.dot(hurr_dis,hurr_dis)*hurr_dis
            d = 111*np.linalg.norm(point - start - proj_point)
            d = np.sign((point - start - proj_point)[0])*d
        coriolis = 4*3.141592653 / 24 * np.sin(np.pi/180.0*point[1])
        if (d < rmw):
            local_wind = intensity
        else:
            local_wind = 2*d*(intensity*rmw+0.5*coriolis*rmw*rmw)/(rmw*rmw+d*d)-(d*coriolis/2)
        windspeeds[idx] = local_wind
        distances[idx] = d
    return windspeeds, distances

def calc_hurricane_hazard_variable(distances, hurricane_speed, radius):
    return np.sqrt(radius*radius - distances) / hurricane_speed

def calc_irma_model(df):
    return calc_hurricane_distance_windspeeds(df,np.array([-81.48,24.5]),np.array([-81.6,25.6]),214,18.5)

def calc_wilma_model(df):
    return calc_hurricane_distance_windspeeds(df,np.array([-83.1,25]),np.array([-81,26.1]),203,30)

def calc_georges_model(df):
    return calc_hurricane_distance_windspeeds(df,np.array([-81.3,23.9]), np.array([-82.4,24.6]),150,30)

def calc_maria_model(df):
    return calc_hurricane_distance_windspeeds(df,np.array([-65.8,18.1]),np.array([-67,18.5]),214,27.8)

def calc_wind_obvs(df, path_to_csv, wind_metric, windspeed_name, wind_attack_name):
    hurr_data = pd.read_csv(path_to_csv)
    num_buoys = len(set(hurr_data['Location']))
    df[wind_attack_name] = 0
    site_max = np.zeros(num_buoys)
    for i in range(num_buoys):
        location = hurr_data.loc[hurr_data['Hour']==0,'Location']
        location = location.iloc[i]
        site_max[i] = np.max(hurr_data.loc[hurr_data['Location'] == location,'Windspeed'])
    for i in df.index:
        site = df.loc[i,['Longitude','Latitude']]
        shore_direction = df.loc[i,'Shore Projection']
        windspeed = df.loc[i, windspeed_name]
        if (len(site.shape)>1):
            windspeed = windspeed.iloc[0]
            shore_direction = np.array(shore_direction.iloc[0])
            site = site.iloc[0]
        shore_direction = shore_direction - site
        shore_direction = shore_direction / np.linalg.norm(shore_direction)
        for t in range(len(set(hurr_data['Hour']))):
            observations = hurr_data[hurr_data['Hour']==t]
            point = np.tile(site,(num_buoys,1)).astype(float)
            distances = circle_dist(observations['lonDD'],  observations['latDD'], point[:,0], point[:,1])
            if (wind_metric == 'MIXED NORMAL' or wind_metric == 'MIXED AVERAGE'):
                fractions = np.multiply(np.divide(observations['Windspeed'].values,site_max), [np.sin((270-observations['Direction'].values)*3.14159/180),np.cos((270-observations['Direction'].values)*3.14159/180)])
                windy =  windspeed * (np.dot(fractions[0,:],np.reciprocal(distances)) / sum(np.reciprocal(distances)))
                windx = windspeed * (np.dot(fractions[1,:],np.reciprocal(distances)) / sum(np.reciprocal(distances)))
            else:
                fractions = np.multiply(observations['Windspeed'].values, [np.sin((270-observations['Direction'].values)*3.14159/180),np.cos((270-observations['Direction'].values)*3.14159/180)])
                windy =  (np.dot(fractions[0,:],np.reciprocal(distances)) / sum(np.reciprocal(distances)))
                windx =(np.dot(fractions[1,:],np.reciprocal(distances)) / sum(np.reciprocal(distances)))
            mag = (np.dot([shore_direction[0], shore_direction[1]], [windx, windy]))
            if (wind_metric == 'AVERAGE' or wind_metric == 'MIXED AVERAGE'):
                df.loc[i,wind_attack_name] = df.loc[i,wind_attack_name] + mag
                continue
            if (len(df.loc[i,wind_attack_name].shape)>0):
                df.loc[i,wind_attack_name] = max(np.max(df.loc[i,wind_attack_name]),mag)
            else:
                df.loc[i,wind_attack_name] = max(df.loc[i,wind_attack_name],mag)
    if (wind_metric == 'AVERAGE' or wind_metric == 'MIXED AVERAGE'):
        df[wind_attack_name] = df[wind_attack_name] / len(set(hurr_data['Hour']))
    return df
        
def read_fknms_data(path_to_data, path_to_stations):
    fk_time_series = pd.read_csv(path_to_data)
    floridakeys_stations =  pd.read_csv(path_to_stations)
    floridakeys_stations = floridakeys_stations.set_index('stationid')
    floridakeys_stations['Longitude'] = floridakeys_stations['lonDD']
    floridakeys_stations['Latitude'] = floridakeys_stations['latDD']
    floridakeys_stations['Depth'] = floridakeys_stations['offshoreDepth']
    floridakeys_stations = floridakeys_stations.drop(columns=['latDD','lonDD'])
    stations_p_change = pd.DataFrame()
    for station in set(fk_time_series['Station_ID']):
        station_time_series = fk_time_series[fk_time_series['Station_ID']==station].set_index('Sample_Year')
        station_time_series = station_time_series[['Macroalgae','Stony coral','Octocoral']].dropna()
        station_p_change = pd.DataFrame()
        for i in range(1,len(station_time_series)):
            ma_change = np.nan
            sc_change = np.nan
            oc_change = np.nan
            tc_change = np.nan
            if (station_time_series.index[i]-station_time_series.index[i-1] == 1):
                ma_change = (2*(station_time_series.iloc[i]['Macroalgae']-station_time_series.iloc[i-1]['Macroalgae'])
                             /(station_time_series.iloc[i-1]['Macroalgae']+station_time_series.iloc[i]['Macroalgae'])
                             if (station_time_series.iloc[i-1]['Macroalgae'] > 0.01) else 
                             np.nan)
                sc_change = (2*(station_time_series.iloc[i]['Stony coral']-station_time_series.iloc[i-1]['Stony coral'])/
                             (station_time_series.iloc[i-1]['Stony coral']+station_time_series.iloc[i]['Stony coral'])
                             if (station_time_series.iloc[i-1]['Stony coral']+station_time_series.iloc[i]['Stony coral'] > 0.05) else 
                             np.nan)
                oc_change = (2*(station_time_series.iloc[i]['Octocoral']-station_time_series.iloc[i-1]['Octocoral'])
                             /(station_time_series.iloc[i-1]['Octocoral']+station_time_series.iloc[i]['Octocoral'])
                             if (station_time_series.iloc[i-1]['Octocoral'] > 0.02) else 
                             np.nan)
                tc_change = (2*(station_time_series.iloc[i]['Octocoral']+station_time_series.iloc[i]['Stony coral']
                             -station_time_series.iloc[i-1]['Octocoral']-station_time_series.iloc[i-1]['Stony coral'])
                             /(station_time_series.iloc[i-1]['Octocoral']+station_time_series.iloc[i]['Octocoral']
                             +station_time_series.iloc[i-1]['Stony coral']+station_time_series.iloc[i]['Stony coral'])
                             if (station_time_series.iloc[i-1]['Octocoral']+station_time_series.iloc[i-1]['Stony coral'] > 0.05) else 
                             np.nan)
            
            station_p_change = station_p_change.append(pd.Series([station_time_series.index[i], 
                                     ma_change, sc_change, oc_change, tc_change,
                                     station_time_series.iloc[i]['Stony coral'], station_time_series.iloc[i]['Octocoral']],name=station))
        station_p_change = station_p_change.reset_index()
        station_p_change.columns = list(range(len(station_p_change.columns)))
        stations_p_change = stations_p_change.append(station_p_change)
    coral_data = dict();
    name = ['Stony coral', 'Octocoral', 'Total coral','Stony coral density', 'Octocoral density']
    for j in range(3,8):
        stony_coral_change = stations_p_change.pivot(0,1,j)
        other = stony_coral_change.copy()
        other['Site_Code'] = floridakeys_stations['Site_Code']
        other = other.groupby('Site_Code').mean()
        other['variance'] = other.var(axis=1)
        other['count'] = np.maximum(other.count(axis=1) - 1,0)
        stony_coral_change_count = stony_coral_change.count(axis=1)
        stony_coral_change['variance'] = stony_coral_change.var(axis=1)
        stony_coral_change['count'] = stony_coral_change_count
        for i in stony_coral_change.index:
            row = other.loc[floridakeys_stations[floridakeys_stations.index==i]['Site_Code']]
            if (len(row)<1):
                continue
            stony_coral_change.loc[i,'variance'] = row['variance'][0]
            stony_coral_change.loc[i,'count'] = row['count'][0]
        coral_data[name[j-3]] = stony_coral_change
    return coral_data, floridakeys_stations